/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * This tool is a tool that display a tree of the current selected program.
 */
Ext.define('Ametys.plugins.odf.tree.AbstractODFTreeTool', {
	extend: "Ametys.plugins.contentstree.ContentsTreeTool",
    
    /**
     * @property {Object[]} _indicators See {@link #cfg-indicators}.
     * @private
     */
    
    /**
     * @cfg {Boolean} [enableValidationCheck=false] Set to 'true' to enable the global validation state on the root content
     */
    /**
     * @property {Boolean} _enableValidationCheck See {@link #cfg-enableValidationCheck}.
     * @private
     */
    /**
     * @cfg {Boolean} [checkValidationAtStartup=false] Set to 'true' to start the global validation check at startup. Use only if {@link #cfg-enableValidationCheck} is true.
     */
    /**
     * @property {Boolean} _checkValidationAtStartup See {@link #cfg-checkValidationAtStartup}.
     * @private
     */
    
    constructor: function(config)
    {
        this._enableValidationCheck = config.enableValidationCheck == true;
        this._checkValidationAtStartup = this._enableValidationCheck && config.checkValidationAtStartup == true;
        
        this.callParent(arguments);
        
        this._indicators = this._getIndicators();
        
        Ametys.message.MessageBus.on(Ametys.message.Message.WORKFLOW_CHANGED, this._onWorkflowChanged, this);
    },
    
    /**
     * @protected
     * Create the tree panel
     * @return {Ext.tree.Panel} the tree panel
     */
    _createTree: function()
    {
        return Ext.create('Ametys.plugins.odf.tree.ODFContentsTreePanel', this._getTreeConfig());
    },
    
    _getTreeConfig: function()
    {
        var treeCfg = this.callParent(arguments);
        return Ext.apply(treeCfg, {
            stateful: true,
            stateId: this.getId() + '$tree',
            indicators: this._getIndicators(),
            enableValidationCheck: this._enableValidationCheck
        });
    },
    
    /**
     * Listener when something has been modified on the workflow
     * @param {Ametys.message.Message} [message] The bus message. Can be null to get the last selection message
     * @private
     */
    _onWorkflowChanged: function(message)
    {
        this._onModified(message);
    },
    
    /**
     * Set the content root node
     * @param {String} contentId The id of root content
     * @param {String} pathToSelect The path to select
     */
    _setContentRootNode: function(contentId, pathToSelect)
    {
        if (this._enableValidationCheck)
        {
            // reset the precede validation state
            this._treePanel.resetValidationState();
        }
        
        this._treePanel.setContentRootNode(
            contentId, 
            this._getRootNodeInfo(),
            Ext.bind(this._setContentRootNodeCb, this, [pathToSelect])
        );
    },
    
    _setContentRootNodeCb: function(pathToSelect)
    {
        this._refreshed(pathToSelect);
        
        if (this._checkValidationAtStartup)
        {
            this._treePanel.checkValidationState();
        }
    },
    
    /**
     * @template
     * Get the list of content types to open on double-click
     * @return {String[]} The array of openable content types
     */
    getOpenableContentTypes: function()
    {
        return [];
    },
    
	_onDoubleClick: function(tree, record, item, index, e, eOpts)
    {
    	var contentId = record.get('contentId');
    	var types = record.get('contenttypesIds') || [];
        var openableTypes = this.getOpenableContentTypes();
        
    	if (contentId && Ext.Array.intersect(openableTypes, types).length > 0)
    	{
    		Ametys.tool.ToolsManager.openTool('uitool-content', {id: contentId, mode: 'view'});
    	}
    },
    
    /**
     * @protected
     * Get for the available indicators
     * @return {Object[]} the available indicators
     */
    _getIndicators: function()
    {
        return [
            // Workflow step
            {
                id: 'workflow-step',
                label: "{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_WORKFLOW_STEP}}",
                description: "{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_WORKFLOW_STEP_DESC}}",
                iconGlyph: 'ametysicon-verified9', 
                matchFn: this._hasWorkflowStep,
                applyFn: this._showWorkflowStep
            },
            // Sharing status
            {
                id: 'shared-status',
                iconGlyph: 'odficon-share-filled', 
                label: "{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_SHARED_STATUS}}",
                description: "{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_SHARED_STATUS_DESC}}",
                matchFn: this._isShared,
                applyFn: this._showSharedStatus
            },
			 // CourseList type
            {
                id: 'courseList-type',
                iconGlyph: 'ametysicon-arrow-up-right-curve', 
                label: "{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_COURSELIST_TYPE}}",
                description: "{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_COURSELIST_TYPE_DESC}}",
                matchFn: this._isCourseList,
                applyFn: this._showCourseListType
            },
             // Is not publishable
            {
                id: 'not-publishable',
                iconGlyph: 'ametysicon-eye-blocked', 
                label: "{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_NOT_PUBLISHABLE}}",
                description: "{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_NOT_PUBLISHABLE_DESC}}",
                matchFn: this._isNotPublishable,
                applyFn: this._showNotPublishable
            },
            // Has shared course part
            {
                id: 'has-shared-course-part',
                iconGlyph: 'odficon-course-part-shared', 
                label: "{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_HAS_SHARED_COURSE_PART}}",
                description: "{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_HAS_SHARED_COURSE_PART_DESC}}",
                matchFn: this._hasSharedCoursePart,
                applyFn: this._showSharedCoursePartStatus
            }
        ]
    },
	
	/**
     * @private
     */
    _isNotPublishable: function(node)
    {
        return node.get('isPublishable') == false || node.get('isParentPublishable') == false;
    },
    
    /**
     * @private
     */
    _showNotPublishable: function(node)
    {
        var isNotPublishable = node.get('isPublishable') == false;
        var parentIsNotPublishable = node.get('isParentPublishable') == false;
        if (isNotPublishable)
        {
            return '<span title="{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_NOT_PUBLISHABLE_TOOLTIP}}" class="' + this.indicatorCls + ' ametysicon-eye-blocked"></span>';
        }
        else if (parentIsNotPublishable)
        {
            return '<span title="{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_PARENT_NOT_PUBLISHABLE_TOOLTIP}}" class="' + this.indicatorCls + ' grey-color ametysicon-eye-blocked"></span>';
        }
    },
	
    /**
	 * @private
	 */
    _isCourseList: function(node)
	{
		return node.get('courseList-type') != undefined;
	},
	
	/**
	 * @private
	 */
    _showCourseListType: function(node)
	{
		var courseListType = node.get('courseList-type')
        
        switch(node.get('courseList-type'))
        {
            case 'CHOICE':
            case 'OPTIONAL':
        }
		
		if (courseListType == 'CHOICE')
		{
            return '<span title="{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_COURSELIST_CHOICE}}" class="' + this.indicatorCls + ' ametysicon-arrow-up-right-curve"></span>';
		}
		else if (courseListType == 'OPTIONAL')
		{
            return '<span title="{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_COURSELIST_OPTIONAL}}" class="' + this.indicatorCls + ' ametysicon-sign-question"></span>';
		}
	},
	
	/**
     * @private
     */
    _hasWorkflowStep: function(node)
    {
        return node.get('workflowStep') != undefined;
    },
    
    /**
     * @private
     */
    _showWorkflowStep: function(node)
    {
        var workflowStep = node.get('workflowStep'); 
        switch (workflowStep.stepId)
        {
            case 3:
                return '<span title="' + workflowStep.name + '" class="' + this.indicatorCls + ' green-color ametysicon-sign-raw-check"></span>';
            case 2:
                return '<span title="' + workflowStep.name + '" class="' + this.indicatorCls + ' orange-color ametysicon-body-part-finger"></span>';
            default:
                return '<span title="' + workflowStep.name + '" class="' + this.indicatorCls + ' orange-color ametysicon-edit45"></span>';
        }                
    },
    
    /**
     * @private
     */
    _isShared: function(node)
    {
        return node.get('shared') || node.get('isParentShared');
    },
    
    /**
     * @private
     */
    _showSharedStatus: function(node)
    {
        // Add shared indicator
        if (node.get('shared'))
        {
            return '<span title="{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_SHARED}}" class="' + this.indicatorCls + ' odficon-share-filled"></span>';
        }
        else
        {
            return '<span title="{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_PARENT_SHARED}}" class="' + this.indicatorCls + ' odficon-share-outline"></span>';
        }
    },
    
    _hasSharedCoursePart: function(node)
    {
        return node.get('hasSharedCoursePart');
    },
    
    _showSharedCoursePartStatus: function(node)
    {
        return `<span title="{{i18n PLUGINS_ODF_CONTENTS_TREE_INDICATORS_HAS_SHARED_COURSE_PART}}" class="${this.indicatorCls} odficon-course-part-shared"></span>`;
    }
});