/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.odf.catalog;

import java.util.List;
import java.util.Map;

import org.ametys.cms.repository.Content;
import org.ametys.odf.program.Program;

/**
 * Interface to update contents after a copy of ODF catalog
 */
public interface CopyCatalogUpdater
{
    /**
     * Update content during copy of a catalog. Implements this method when you do not need to wait for the full copy of the catalog.
     * @param initialCatalogName the name of initial catalog
     * @param newCatalogName the name of new catalog
     * @param initialProgram the initial program
     * @param createdProgram the created program
     */
    public void updateContent (String initialCatalogName, String newCatalogName, Program initialProgram, Program createdProgram);
    
    /**
     * Update contents after the full copy of a catalog. Implements this method when you need to do some stuff when all contents have been copied.
     * @param initialCatalogName the name of initial catalog
     * @param newCatalogName the name of new catalog
     * @param copiedPrograms the id of initial programs with their copied content
     * @param copiedSubPrograms the id of initial subprograms with their copied content
     * @param copiedContainers the id of initial containers with their copied content
     * @param copiedCourseLists the id of initial course lists with their copied content
     * @param copiedCourses the id of initial courses with their copied content
     * @param copiedCourseParts the id of initial course parts with their copied content
     */
    public void updateContents (String initialCatalogName, String newCatalogName, Map<String, String> copiedPrograms, Map<String, String> copiedSubPrograms, Map<String, String> copiedContainers, Map<String, String> copiedCourseLists, Map<String, String> copiedCourses, Map<String, String> copiedCourseParts);
    
    /**
     * Get the additional contents of the catalog
     * @param catalogName the catalog name
     * @return additional contents managed by the updater
     */
    public default List<Content> getAdditionalContents(String catalogName)
    {
        return List.of();
    }
    
    /**
     * Copy additional contents from the initial to the new catalog
     * @param initialCatalogName the name of initial catalog
     * @param newCatalogName the name of new catalog
     */
    public default void copyAdditionalContents(String initialCatalogName, String newCatalogName)
    {
        // Nothing to do by default
    }
}
