/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.odf.content;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Optional;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.odf.NoLiveVersionException;
import org.ametys.odf.ODFHelper;
import org.ametys.odf.ProgramItem;
import org.ametys.odf.course.Course;
import org.ametys.odf.courselist.CourseList;
import org.ametys.odf.coursepart.CoursePart;
import org.ametys.odf.enumeration.OdfReferenceTableEntry;
import org.ametys.odf.enumeration.OdfReferenceTableHelper;
import org.ametys.odf.program.AbstractProgram;
import org.ametys.odf.program.Container;
import org.ametys.odf.program.Program;
import org.ametys.odf.program.SubProgram;
import org.ametys.odf.skill.ODFSkillsHelper;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.jcr.DefaultAmetysObject;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.DataContext;

/**
 * SAX the structure (ie. the child program items) of a {@link ProgramItem}
 *
 */
public class ProgramItemStructureGenerator extends ServiceableGenerator
{
    private static final Set<String> __ALLOWED_VIEW_NAMES = Set.of("main", "pdf");
    
    /** The ODF helper */
    protected ODFHelper _odfHelper;
    /** Helper for ODF reference table */
    protected OdfReferenceTableHelper _odfReferenceTableHelper;
    /** The content types helper */
    protected ContentTypesHelper _cTypesHelper;
    /** The ODF skills helper */
    protected ODFSkillsHelper _odfSkillsHelper;
    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _odfHelper = (ODFHelper) smanager.lookup(ODFHelper.ROLE);
        _odfReferenceTableHelper = (OdfReferenceTableHelper) smanager.lookup(OdfReferenceTableHelper.ROLE);
        _cTypesHelper = (ContentTypesHelper) smanager.lookup(ContentTypesHelper.ROLE);
        _odfSkillsHelper = (ODFSkillsHelper) smanager.lookup(ODFSkillsHelper.ROLE);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
    }
    
    public void generate() throws IOException, SAXException, ProcessingException
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        Content content = (Content) request.getAttribute(Content.class.getName());
        
        if (content == null)
        {
            String contentId = parameters.getParameter("contentId", null);
            if (StringUtils.isBlank(contentId))
            {
                throw new IllegalArgumentException("Content is missing in request attribute or parameters");
            }
            content = _resolver.resolveById(contentId);
        }

        String viewName = parameters.getParameter("viewName", StringUtils.EMPTY);
        String fallbackViewName = parameters.getParameter("fallbackViewName", StringUtils.EMPTY);
        
        View view = _cTypesHelper.getViewWithFallback(viewName, fallbackViewName, content);
        
        contentHandler.startDocument();
        
        if (view != null && __ALLOWED_VIEW_NAMES.contains(view.getName()))
        {
            if (content instanceof ProgramItem)
            {
                XMLUtils.startElement(contentHandler, "structure");
                saxProgramItem((ProgramItem) content); 
                XMLUtils.endElement(contentHandler, "structure");
            }
            else
            {
                getLogger().warn("Cannot get the structure of a non program item '" + content.getId() + "'");
            }
        }
        
        contentHandler.endDocument();
    }
        
    /**
     * SAX a program item with its child program items
     * @param programItem the program item
     * @throws SAXException if an error occurs while saxing
     */
    protected void saxProgramItem(ProgramItem programItem) throws SAXException
    {
        if (programItem instanceof Program)
        {
            saxProgram((Program) programItem);
        }
        else if (programItem instanceof SubProgram)
        {
            saxSubProgram((SubProgram) programItem);
        }
        else if (programItem instanceof Container)
        {
            saxContainer((Container) programItem);
        }
        else if (programItem instanceof CourseList)
        {
            saxCourseList((CourseList) programItem);
        }
        else if (programItem instanceof Course)
        {
            saxCourse((Course) programItem);
        }
    }
    
    /**
     * SAX the child program items
     * @param programItem the program item
     * @throws SAXException if an error occurs while saxing
     */
    protected void saxChildProgramItems(ProgramItem programItem) throws SAXException
    {
        List<ProgramItem> childProgramItems = _odfHelper.getChildProgramItems(programItem);
        for (ProgramItem childProgramItem : childProgramItems)
        {
            try
            {
                _odfHelper.switchToLiveVersionIfNeeded((DefaultAmetysObject) childProgramItem);
                saxProgramItem(childProgramItem);
            }
            catch (NoLiveVersionException e) 
            {
                // Just ignore the program item
            }
        }
    }
        
    /**
     * SAX a program
     * @param program the subprogram to SAX
     * @throws SAXException if an error occurs
     */
    protected void saxProgram(Program program) throws SAXException
    {
        AttributesImpl attrs = new AttributesImpl();
        _saxCommonAttributes(program, attrs);
        
        XMLUtils.startElement(contentHandler, "program", attrs);
        
        XMLUtils.startElement(contentHandler, "attributes");
        _saxStructureViewIfExists(program);
        XMLUtils.endElement(contentHandler, "attributes");
        
        saxChildProgramItems(program);
        XMLUtils.endElement(contentHandler, "program");
    }
    
    /**
     * SAX a subprogram
     * @param subProgram the subprogram to SAX
     * @throws SAXException if an error occurs
     */
    protected void saxSubProgram(SubProgram subProgram) throws SAXException
    {
        AttributesImpl attrs = new AttributesImpl();
        _saxCommonAttributes(subProgram, attrs);
        
        XMLUtils.startElement(contentHandler, "subprogram", attrs);
        
        XMLUtils.startElement(contentHandler, "attributes");
        _saxReferenceTableItem(subProgram.getEcts(), AbstractProgram.ECTS, subProgram.getLanguage());
        _saxStructureViewIfExists(subProgram);
        XMLUtils.endElement(contentHandler, "attributes");
        
        saxChildProgramItems(subProgram);
        
        XMLUtils.endElement(contentHandler, "subprogram");
    }
    
    /**
     * SAX a container
     * @param container the container to SAX
     * @throws SAXException if an error occurs while saxing
     */
    protected void saxContainer(Container container) throws SAXException
    {
        AttributesImpl attrs = new AttributesImpl();
        _saxCommonAttributes(container, attrs);
        
        XMLUtils.startElement(contentHandler, "container", attrs);
        
        XMLUtils.startElement(contentHandler, "attributes");
        _saxReferenceTableItem(container.getNature(), Container.NATURE, container.getLanguage());
        _saxReferenceTableItem(container.getPeriod(), Container.PERIOD, container.getLanguage());
        
        _saxStructureViewIfExists(container);
        
        XMLUtils.endElement(contentHandler, "attributes");
        
        saxChildProgramItems(container);
        
        XMLUtils.endElement(contentHandler, "container");
    }
    
    /**
     * SAX a course list
     * @param cl the course list to SAX
     * @throws SAXException if an error occurs while saxing
     */
    protected void saxCourseList(CourseList cl) throws SAXException
    {
        AttributesImpl attrs = new AttributesImpl();
        _saxCommonAttributes(cl, attrs);
        
        XMLUtils.startElement(contentHandler, "courselist", attrs);
        
        XMLUtils.startElement(contentHandler, "attributes");
        _saxStructureViewIfExists(cl);
        XMLUtils.endElement(contentHandler, "attributes");
        
        saxChildProgramItems(cl);
        
        XMLUtils.endElement(contentHandler, "courselist");
    }
    
    /**
     * SAX a course
     * @param course the container to SAX
     * @throws SAXException if an error occurs while saxing
     */
    protected void saxCourse(Course course) throws SAXException
    {
        AttributesImpl attrs = new AttributesImpl();
        _saxCommonAttributes(course, attrs);
        
        XMLUtils.startElement(contentHandler, "course", attrs);
        
        XMLUtils.startElement(contentHandler, "attributes");
        _saxReferenceTableItem(course.getCourseType(), Course.COURSE_TYPE, course.getLanguage());
        _saxStructureViewIfExists(course);
        XMLUtils.endElement(contentHandler, "attributes");
        
        saxChildProgramItems(course);
        
        saxCourseParts(course);
        
        XMLUtils.endElement(contentHandler, "course");
    }
    
    /**
     * SAX a course part
     * @param course The course
     * @throws SAXException if an error occurs
     */
    protected void saxCourseParts(Course course) throws SAXException
    {
        List<CoursePart> courseParts = course.getCourseParts();
        
        double totalHours = 0;
        List<CoursePart> liveCourseParts = new ArrayList<>();
        for (CoursePart coursePart : courseParts)
        {
            try
            {
                _odfHelper.switchToLiveVersionIfNeeded(coursePart);
                totalHours += coursePart.getNumberOfHours();
                liveCourseParts.add(coursePart);
            }
            catch (NoLiveVersionException e) 
            {
                getLogger().warn("Some hours of " + course.toString() + " are not added because the course part " + coursePart.toString() + " does not have a live version.");
            }
        }
        
        AttributesImpl attrs = new AttributesImpl();
        attrs.addCDATAAttribute("totalHours", String.valueOf(totalHours));
        XMLUtils.startElement(contentHandler, "courseparts", attrs);

        for (CoursePart coursePart : liveCourseParts)
        {
            saxCoursePart(coursePart);
        }
        
        XMLUtils.endElement(contentHandler, "courseparts");
    }
    
    /**
     * SAX a course part
     * @param coursePart The course part to SAX
     * @throws SAXException if an error occurs
     */
    protected void saxCoursePart(CoursePart coursePart) throws SAXException
    {
        AttributesImpl attrs = new AttributesImpl();
        attrs.addCDATAAttribute("id", coursePart.getId());
        attrs.addCDATAAttribute("title", coursePart.getTitle());
        _addAttrIfNotEmpty(attrs, "code", coursePart.getCode());

        XMLUtils.startElement(contentHandler, "coursepart", attrs);
        
        XMLUtils.startElement(contentHandler, "attributes");
        _saxReferenceTableItem(coursePart.getNature(), CoursePart.NATURE, coursePart.getLanguage());
        
        _saxStructureViewIfExists(coursePart);
        
        XMLUtils.endElement(contentHandler, "attributes");
        
        XMLUtils.endElement(contentHandler, "coursepart");
    }
    
    /**
     * SAX the 'structure' view if exists
     * @param content the content
     * @throws SAXException if an error occurs
     */
    protected void _saxStructureViewIfExists(Content content) throws SAXException
    {
        View view = _cTypesHelper.getView("structure", content.getTypes(), content.getMixinTypes());
        if (view != null)
        {
            try
            {
                content.dataToSAX(contentHandler, view, DataContext.newInstance().withLocale(new Locale(content.getLanguage())).withEmptyValues(false));
            }
            catch (BadItemTypeException | AmetysRepositoryException e)
            {
                throw new SAXException("Fail to sax the 'structure' view for content " + content.getId(), e);
            }
        }
    }
    
    /**
     * SAX the common attributes for program item
     * @param programItem the program item
     * @param attrs the attributes
     */
    protected void _saxCommonAttributes(ProgramItem programItem, AttributesImpl attrs)
    {
        attrs.addCDATAAttribute("title", ((Content) programItem).getTitle());
        attrs.addCDATAAttribute("id", programItem.getId());
        attrs.addCDATAAttribute("code", programItem.getCode());
        attrs.addCDATAAttribute("name", programItem.getName());
    
        boolean excludedFromSkills = _odfSkillsHelper.isExcluded(programItem);
        if (excludedFromSkills)
        {
            attrs.addCDATAAttribute("excludedFromSkills", String.valueOf(excludedFromSkills));
        }
    }
    
    /**
     * SAX the item of a reference table
     * @param itemId the item id
     * @param tagName the tag name
     * @param lang the language to use
     * @throws SAXException if an error occurs while saxing
     */
    protected void _saxReferenceTableItem(String itemId, String tagName, String lang) throws SAXException
    {
        OdfReferenceTableEntry item = Optional.ofNullable(itemId)
                                              .filter(StringUtils::isNotEmpty)
                                              .map(_odfReferenceTableHelper::getItem)
                                              .orElse(null);
        
        if (item != null)
        {
            AttributesImpl attrs = new AttributesImpl();
            attrs.addCDATAAttribute("id", item.getId());
            _addAttrIfNotEmpty(attrs, "code", item.getCode());
            
            XMLUtils.createElement(contentHandler, tagName, attrs, item.getLabel(lang));
        }
    }
    
    /**
     * Add an attribute if its not null or empty.
     * @param attrs The attributes
     * @param attrName The attribute name
     * @param attrValue The attribute value
     */
    protected void _addAttrIfNotEmpty(AttributesImpl attrs, String attrName, String attrValue)
    {
        if (StringUtils.isNotEmpty(attrValue))
        {
            attrs.addCDATAAttribute(attrName, attrValue);
        }
    }

}
