/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.odf.program;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.cms.workflow.ContentWorkflowHelper;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.ui.Callable;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.odf.ODFHelper;
import org.ametys.odf.observation.OdfObservationConstants;
import org.ametys.odf.translation.TranslationHelper;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.UnknownAmetysObjectException;

/**
 * DAO for manipulating ODF programs.
 *
 */
public class ProgramDAO implements Serviceable, Component
{
    /** The Avalon role */
    public static final String ROLE = ProgramDAO.class.getName();
    
    /** The ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** Observer manager. */
    protected ObservationManager _observationManager;
    /** The current user provider. */
    protected CurrentUserProvider _currentUserProvider;
    /** ODF helper */
    protected ODFHelper _odfHelper;
    /** The content workflow helper */
    protected ContentWorkflowHelper _contentWorkflowHelper;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _observationManager = (ObservationManager) manager.lookup(ObservationManager.ROLE);
        _currentUserProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
        _odfHelper = (ODFHelper) manager.lookup(ODFHelper.ROLE);
        _contentWorkflowHelper = (ContentWorkflowHelper) manager.lookup(ContentWorkflowHelper.ROLE);
    }
    
    /**
     * Translates a given program in a language
     * @param contentId The program id
     * @param language The language in which to translate
     * @param fullCopy True if full copy
     * @return A Map with id of translated program or an error message
     * @throws Exception if an error occurs
     */
    @Callable
    public Map<String, String> translateProgram (String contentId, String language, boolean fullCopy) throws Exception
    {
        Map<String, String> result = new HashMap<>();
        
        Program program = _resolver.resolveById(contentId);
        
        if (program.getLanguage().equals(language))
        {
            result.put("error", "same-language");
            return result;
        }
        
        Map<String, String> translatedContents = new HashMap<>();
        
        String translatedProgramId = null;
        
        // Get existing translations
        Map<String, String> translations = TranslationHelper.getTranslations(program);
        
        if (!translations.containsKey(language))
        {
            ModifiableContent translatedProgram = _odfHelper.copyProgramItem(program, null, language, null, fullCopy, translatedContents, translatedContents, translatedContents, translatedContents, translatedContents, translatedContents);
            translatedProgramId = translatedProgram.getId();
        }
        else
        {
            translatedProgramId = translations.get(language);
            // Check if content exists
            try
            {
                _resolver.resolveById(translatedProgramId);
                result.put("error", "already-exists");
            }
            catch (UnknownAmetysObjectException e) 
            {
                ModifiableContent translatedProgram = _odfHelper.copyProgramItem(program, null, language, null, fullCopy, translatedContents, translatedContents, translatedContents, translatedContents, translatedContents, translatedContents);
                translatedProgramId = translatedProgram.getId();
            }
        }
        
        linkTranslations(translatedContents);
        
        // Add the workflow step, send notifications and extract outgoing references
        for (String newContentId : translatedContents.values())
        {
            WorkflowAwareContent newContent = _resolver.resolveById(newContentId);
            _contentWorkflowHelper.doAction(newContent, getCopyActionId());
        }
        
        result.put("translated-program-id", translatedProgramId);
        
        return result;
    }
    
    /**
     * Store links to the other translations for all copied contents
     * @param translatedContents The translated contents
     */
    protected void linkTranslations(Map<String, String> translatedContents)
    {
        for (Entry<String, String> entry : translatedContents.entrySet())
        {
            ModifiableContent originalContent = _resolver.resolveById(entry.getKey());
            ModifiableContent translatedContent = _resolver.resolveById(entry.getValue());
            
            linkTranslations(originalContent, translatedContent);
            
            originalContent.saveChanges();
            translatedContent.saveChanges();
            
            Map<String, Object> eventParams = new HashMap<>();
            eventParams.put(ObservationConstants.ARGS_CONTENT, originalContent);
            
            _observationManager.notify(new Event(OdfObservationConstants.ODF_CONTENT_TRANSLATED, _getCurrentUser(), eventParams));
        }
    }
    
    /**
     * Store links to the other translations in all the translated objects.
     * @param originalContent The original content
     * @param translatedContent The translated content
     */
    protected void linkTranslations(ModifiableContent originalContent, ModifiableContent translatedContent)
    {
        Map<String, String> translations = TranslationHelper.getTranslations(originalContent);
        
        // Add the original and translated references.
        translations.put(originalContent.getLanguage(), originalContent.getId());
        translations.put(translatedContent.getLanguage(), translatedContent.getId());
        
        for (String contentId : translations.values())
        {
            ModifiableContent content = _resolver.resolveById(contentId);
            
            Map<String, String> otherTranslations = new HashMap<>(translations);
            otherTranslations.remove(content.getLanguage());
            
            TranslationHelper.setTranslations(content, otherTranslations);
        }
    }
    
    /**
     * Provides the current user.
     * @return the identity which cannot be <code>null</code>.
     */
    protected UserIdentity _getCurrentUser()
    {
        return _currentUserProvider.getUser();
    }
    
    /**
     * Get the workflow action id for copy.
     * @return The workflow action id
     */
    protected int getCopyActionId()
    {
        return 210;
    }
}
