/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.odf.rights;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.repository.Content;
import org.ametys.core.right.AccessController;
import org.ametys.core.right.RightsException;
import org.ametys.odf.ODFHelper;
import org.ametys.odf.ProgramItem;
import org.ametys.odf.coursepart.CoursePart;
import org.ametys.odf.orgunit.OrgUnit;
import org.ametys.odf.person.Person;
import org.ametys.odf.rights.ODFRightHelper.PermissionContext;
import org.ametys.plugins.core.impl.right.AbstractHierarchicalWithPermissionContextAccessController;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.i18n.I18nizableTextParameter;

/**
 * {@link AccessController} for a ODF {@link Content}
 */
public class ODFContentHierarchicalAccessController extends AbstractHierarchicalWithPermissionContextAccessController<AmetysObject, PermissionContext> implements Contextualizable
{
    /** The helper for root content */
    protected ODFHelper _odfHelper;
    /** The helper for contents */
    protected ContentHelper _contentHelper;
    /** The avalon context */
    protected Context _context;
    /** Ametys Object Resolver */
    protected AmetysObjectResolver _resolver;
    /** Th eright helper for ODF contents */
    protected ODFRightHelper _odfRightHelper;


    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _odfHelper = (ODFHelper) manager.lookup(ODFHelper.ROLE);
        _odfRightHelper = (ODFRightHelper) manager.lookup(ODFRightHelper.ROLE);
        _contentHelper = (ContentHelper) manager.lookup(ContentHelper.ROLE);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    public boolean isSupported(Object object)
    {
        return object instanceof ProgramItem || object instanceof OrgUnit || object instanceof Person || object instanceof CoursePart;
    }
    
    @Override
    protected Set<AmetysObject> _getParents(AmetysObject object, PermissionContext permissionCtx)
    {
        if (!(object instanceof Content))
        {
            return null;
        }
        
        return _odfRightHelper.getParents((Content) object, permissionCtx);
    }
    
    @Override
    protected PermissionContext _getPermissionContext(AmetysObject initialObject)
    {
        return new PermissionContext((Content) initialObject);
    }
    
    @Override
    protected Set< ? extends Object> _convertWorkspaceToRootRightContexts(Set<Object> workspacesContexts)
    {
        if (workspacesContexts.contains("/cms"))
        {
            return Collections.singleton(_odfHelper.getRootContent());
        }
        return null;
    }
    
    @Override
    protected I18nizableText _getObjectLabel(Object object) throws RightsException
    {
        if (object instanceof Content content)
        {
            Map<String, I18nizableTextParameter> params = new HashMap<>();
            params.put("title", new I18nizableText(content.getTitle()));
            if (object instanceof ProgramItem item)
            {
                params.put("code", new I18nizableText(item.getCode()));
                return new I18nizableText("plugin.odf", "PLUGINS_ODF_ODF_CONTENT_ACCESS_CONTROLLER_PROGRAM_ITEM_CONTEXT_LABEL", params);
            }
            else if (object instanceof CoursePart part)
            {
                params.put("code", new I18nizableText(part.getCode()));
                return new I18nizableText("plugin.odf", "PLUGINS_ODF_ODF_CONTENT_ACCESS_CONTROLLER_COURSE_PART_CONTEXT_LABEL", params);
            }
            else if (object instanceof OrgUnit orgunit)
            {
                params.put("code", new I18nizableText(orgunit.getUAICode()));
                return new I18nizableText("plugin.odf", "PLUGINS_ODF_ODF_CONTENT_ACCESS_CONTROLLER_ORGUNIT_CONTEXT_LABEL", params);
            }
            else if (object instanceof Person)
            {
                return new I18nizableText("plugin.odf", "PLUGINS_ODF_ODF_CONTENT_ACCESS_CONTROLLER_PERSON_CONTEXT_LABEL", params);
            }
        }
        else if (object.equals(_odfHelper.getRootContent()))
        {
            return new I18nizableText("plugin.odf", "PLUGINS_ODF_ODF_CONTENT_ACCESS_CONTROLLER_ROOT_CONTEXT_LABEL");
        }
        
        throw new RightsException("Unsupported object " + object.toString());
    }
}
