/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.odf.tree;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.ui.Callable;
import org.ametys.odf.ODFHelper;
import org.ametys.odf.ProgramItem;
import org.ametys.odf.course.Course;
import org.ametys.odf.course.ShareableCourseHelper;
import org.ametys.odf.course.ShareableCourseStatusHelper;
import org.ametys.odf.course.ShareableCourseStatusHelper.ShareableStatus;
import org.ametys.odf.courselist.CourseList;
import org.ametys.odf.coursepart.CoursePart;
import org.ametys.odf.orgunit.OrgUnit;
import org.ametys.plugins.contentstree.ContentsTreeHelper;
import org.ametys.plugins.workflow.repository.WorkflowAwareAmetysObject;
import org.ametys.plugins.workflow.support.WorkflowProvider;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;
import org.ametys.runtime.i18n.I18nizableText;

import com.opensymphony.workflow.loader.StepDescriptor;
import com.opensymphony.workflow.loader.WorkflowDescriptor;

/**
 * Helper ODF contents tree
 *
 */
public class ODFContentsTreeHelper extends ContentsTreeHelper
{
    /** The Avalon role */
    @SuppressWarnings("hiding")
    public static final String ROLE = ODFContentsTreeHelper.class.getName();
 
    /** The ODF helper */
    protected ODFHelper _odfHelper;
    /** The workflow provider */
    protected WorkflowProvider _workflowProvider;
    /** The shareable course status helper */
    protected ShareableCourseStatusHelper _shareableStatusHelper;
    /** The shareable course helper */
    protected ShareableCourseHelper _shareableCourseHelper;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        
        _workflowProvider = (WorkflowProvider) smanager.lookup(WorkflowProvider.ROLE);
        _odfHelper = (ODFHelper) smanager.lookup(ODFHelper.ROLE);
        _shareableCourseHelper = (ShareableCourseHelper) smanager.lookup(ShareableCourseHelper.ROLE);
        _shareableStatusHelper = (ShareableCourseStatusHelper) smanager.lookup(ShareableCourseStatusHelper.ROLE);
    }
    
    @Override
    @Callable
    public Map<String, Object> getRootNodeInformations(String contentId, String treeId)
    {
        Map<String, Object> infos = super.getRootNodeInformations(contentId, treeId);
        infos.put("handleShareableCourse", _shareableCourseHelper.handleShareableCourse());
        return infos;
    }
    
    @Override
    protected Map<String, Object> content2Json(Content content)
    {
        Map<String, Object> infos = super.content2Json(content);
        if (content instanceof ProgramItem programItem)
        {
            boolean isShared = isShared(programItem);
            boolean isSharedByParents = !isShared ? isSharedByParents(programItem) : false;
            infos.put("code", getProgramItemDisplayCode((Content & ProgramItem) content));
            infos.put("workflowStep", getWorkflowStep(content));
            infos.put("isShared", isShared);
            infos.put("isParentShared", isSharedByParents);
            infos.put("isPublishable", programItem.isPublishable());
            infos.put("isParentPublishable", _isParentPublishable(programItem));
            
            if (content instanceof Course)
            {
                infos.put("shareableStatus", getShareableStatus((Course) content));
                infos.put("hasSharedCoursePart", _hasSharedCoursePart((Course) content));
            }
            else if (content instanceof CourseList)
            {
                CourseList courseList = (CourseList) content;
                infos.put("courseList-type", courseList.getType());
            }
        }
        else if (content instanceof OrgUnit)
        {
            infos.put("code", ((OrgUnit) content).getUAICode());
        }
        
        return infos;
    }
    
    private boolean _hasSharedCoursePart(Course content)
    {
        return content.getCourseParts().stream()            // parcourt toutes les heures d'enseignement du cours
            .map(CoursePart::getCourses)                    // récupère la liste des cours lié à chaque heure d'enseignement
            .flatMap(List::stream)                          // parcourt la liste
            .anyMatch(c -> !StringUtils.equals(c.getId(), content.getId()));   // vérifie si un des cours n'est pas le cours actuel
    }

    private boolean _isParentPublishable(ProgramItem content)
    {
        List<ProgramItem> parents = _odfHelper.getParentProgramItems(content);
        if (parents.isEmpty())
        {
            return true;
        }
        
        return parents.stream()
                .filter(ProgramItem::isPublishable)
                .filter(this::_isParentPublishable)
                .findAny()
                .isPresent();
    }
    
    /**
     * Get the ProgramItem code we want to display.
     * @param <T> The type of the program element, should be a subclasse of {@link ProgramItem} and {@link Content}
     * @param programItem The program item
     * @return The code to display
     */
    protected <T extends ProgramItem & Content> Object getProgramItemDisplayCode(T programItem)
    {
        return programItem.getCode();
    }
    
    @Override
    protected boolean isContentMatching(Content content, String value)
    {
        boolean matchTitle = super.isContentMatching(content, value);
        if (!matchTitle)
        {
            if (content instanceof ProgramItem)
            {
                // Try with code
                String code =  StringUtils.stripAccents(((ProgramItem) content).getCode().toLowerCase());
                return code.contains(value);
            }
            else if (content instanceof OrgUnit)
            {
                // Try with code
                String code =  StringUtils.stripAccents(((OrgUnit) content).getUAICode().toLowerCase());
                return code.contains(value);
            }
        }
        
        return matchTitle;
    }
    
    /**
     * Get workflow step information
     * @param content the content the content
     * @return the workflow step information
     */
    protected Map<String, Object> getWorkflowStep(Content content)
    {
        Map<String, Object> workflowInfos = new HashMap<>();
                
        if (content instanceof WorkflowAwareAmetysObject)
        {
            WorkflowAwareContent waContent = (WorkflowAwareContent) content;
            
            long workflowId = waContent.getWorkflowId();
            int currentStepId = Math.toIntExact(waContent.getCurrentStepId());
        
            AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(waContent);
        
            String workflowName = workflow.getWorkflowName(workflowId);
            WorkflowDescriptor workflowDescriptor = workflow.getWorkflowDescriptor(workflowName);
        
            if (workflowDescriptor != null)
            {
                StepDescriptor stepDescriptor = workflowDescriptor.getStep(currentStepId);
                if (stepDescriptor != null)
                {
                    I18nizableText workflowStepName = new I18nizableText("application", stepDescriptor.getName());
                    
                    workflowInfos.put("stepId", currentStepId);
                    workflowInfos.put("name", workflowStepName);
                    
                    String[] icons = new String[] {"small", "medium", "large"};
                    for (String icon : icons)
                    {
                        if ("application".equals(workflowStepName.getCatalogue()))
                        {
                            workflowInfos.put(icon + "Icon", "/plugins/cms/resources_workflow/" + workflowStepName.getKey() + "-" + icon + ".png");
                        }
                        else
                        {
                            String pluginName = workflowStepName.getCatalogue().substring("plugin.".length());
                            workflowInfos.put(icon + "Icon", "/plugins/" + pluginName + "/resources/img/workflow/" + workflowStepName.getKey() + "-" + icon + ".png");
                        }
                    }
                }
            }
        }
        
        return workflowInfos;
    }
    
    /**
     * Get the shared status of this program item
     * @param programItem the program item
     * @return true if the program item belongs to more than one program item
     */
    protected boolean isShared(ProgramItem programItem)
    {
        return _odfHelper.getParentProgramItems(programItem).size() > 1;
    }
    
    /**
     * Check if the parent of a program item is shared 
     * @param programItem the program item
     * @return true if the parent of the program item is shared
     */
    protected boolean isSharedByParents(ProgramItem programItem)
    {
        for (ProgramItem parent : _odfHelper.getParentProgramItems(programItem))
        {
            return _odfHelper.getParentProgramItems(parent).size() > 1 ? true : isSharedByParents(parent);
        }
        return false;
    }

    /**
     * Get the shareable course status as String
     * @param course the course
     * @return the shareable status
     */
    protected String getShareableStatus(Course course)
    {
        if (_shareableCourseHelper.handleShareableCourse())
        {
            ShareableStatus shareableStatus = _shareableStatusHelper.getShareableStatus(course);
            return shareableStatus.name().toLowerCase();
        }
        
        return null;
    }
}
