/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.odf.workflow.copy;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.clientsideelement.SmartContentClientSideElement;
import org.ametys.cms.indexing.solr.SolrIndexHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.workflow.ContentWorkflowHelper;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.ui.Callable;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.JSONUtils;
import org.ametys.odf.ODFHelper;
import org.ametys.odf.ProgramItem;
import org.ametys.odf.course.Course;
import org.ametys.odf.course.ShareableCourseHelper;
import org.ametys.odf.courselist.CourseList;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Client side element for ODF content copy
 * 
 */
public class CopyODFContentClientSideElement extends SmartContentClientSideElement
{
    /** The key to get the duplication mode */
    public static final String DUPLICATION_MODE_KEY = "$duplicationMode";
    
    /** The key to get the parent ProgramPart's id */
    public static final String PARENT_KEY = "$parent";
    
    /** The key to tell if we keep the creation title */
    public static final String KEEP_CREATION_TITLE_KEY = "$keepCreationTitle";
    
    /** The content workflow helper */
    protected ContentWorkflowHelper _contentWorkflowHelper;
    /** The observation manager */
    protected ObservationManager _observationManager;
    /** The ODF helper */
    protected ODFHelper _odfHelper;
    /** The client side element for copy */
    protected CopyContentClientInteraction _copyClientSideInteraction;
    /** JSON utils */
    protected JSONUtils _jsonUtils;
    /** The shareable course helper */
    protected ShareableCourseHelper _shareableCourseHelper;
    /** The Solr index helper */
    protected SolrIndexHelper _solrIndexHelper;
    
    @Override
    public void service(ServiceManager sManager) throws ServiceException
    {
        super.service(sManager);
        _contentWorkflowHelper = (ContentWorkflowHelper) sManager.lookup(ContentWorkflowHelper.ROLE);
        _observationManager = (ObservationManager) sManager.lookup(ObservationManager.ROLE);
        _odfHelper = (ODFHelper) sManager.lookup(ODFHelper.ROLE);
        _copyClientSideInteraction = (CopyContentClientInteraction) sManager.lookup(CopyContentClientInteraction.class.getName());
        _jsonUtils = (JSONUtils) sManager.lookup(JSONUtils.ROLE);
        _shareableCourseHelper = (ShareableCourseHelper) sManager.lookup(ShareableCourseHelper.ROLE);
        _solrIndexHelper = (SolrIndexHelper) sManager.lookup(SolrIndexHelper.ROLE);
    }
    
    /**
     * Determines if a ODF content can be copied and linked to a target content
     * @param copiedContentId The id of copied content
     * @param targetContentId The id of target content
     * @param contextualParameters the contextual parameters
     * @return the result with success to true if copy is available
     */
    @Callable
    public Map<String, Object> canCopyTo(String copiedContentId, String targetContentId, Map<String, Object> contextualParameters)
    {
        Content copiedContent = _resolver.resolveById(copiedContentId);
        Content targetContent = _resolver.resolveById(targetContentId);
           
        Map<String, Object> result = new HashMap<>();
        
        // Put the mode to copy to prevent to check shareable course fields
        contextualParameters.put("mode", "copy");
        
        List<I18nizableText> errors = new ArrayList<>();
        if (!_odfHelper.isRelationCompatible(copiedContent, targetContent, errors, contextualParameters))
        {
            // Invalid target
            result.put("errorMessages", errors);
            result.put("success", false); 
        }
        else
        {
            result.put("success", true); 
        }
        
        return result;
    }
    
    /**
     * Creates a content by copy of another one.<br>
     * Also handle the possible inner duplication depending on the duplication mode for each attribute of type "content". 
     * @param baseContentId The id of content to copy
     * @param newContentTitle The title of content to create
     * @param viewNameToCopy The view name to copy. Can be null
     * @param fallbackViewNameToCopy The fallback view name to use if 'viewNameToCopy' does not exist. Can be null
     * @param viewMode The view type to copy. Can be null
     * @param initActionId The init workflow action id for copy
     * @param editActionId The workflow action for editing content
     * @param duplicationModeAsString the duplication mode
     * @param parentContentId the parent id under which the duplicated content will be created. Can be null
     * @return the copy result
     * @throws Exception if an error occurred during copy
     */
    @Callable
    public Map<String, Object> createContentByCopy(String baseContentId, String newContentTitle, String viewNameToCopy, String fallbackViewNameToCopy, String viewMode, int initActionId, int editActionId, String duplicationModeAsString, String parentContentId) throws Exception
    {
        Map<String, Object> result = new HashMap<>();
        result.put("locked-contents", new ArrayList<>());

        String [] handledEventIds = new String[] {ObservationConstants.EVENT_CONTENT_ADDED, ObservationConstants.EVENT_CONTENT_MODIFIED,  ObservationConstants.EVENT_CONTENT_WORKFLOW_CHANGED};
        try
        {
            _solrIndexHelper.pauseSolrCommitForEvents(handledEventIds);
            
            DuplicationMode duplicationMode = StringUtils.isNotBlank(duplicationModeAsString) ? DuplicationMode.valueOf(duplicationModeAsString.toUpperCase()) : DuplicationMode.SINGLE;
            
            if (!checkBeforeDuplication(baseContentId, parentContentId, duplicationMode, result))
            {
                result.put("check-before-duplication-failed", true);
            }
            else
            {
                Map<String, Object> copyMap = new HashMap<>();
                copyMap.put(DUPLICATION_MODE_KEY, duplicationMode.toString());
                copyMap.put(PARENT_KEY, parentContentId);
                
                String jsonMap = _jsonUtils.convertObjectToJson(copyMap);
                result = _copyClientSideInteraction.createContentByCopy(baseContentId, newContentTitle, jsonMap, viewNameToCopy, fallbackViewNameToCopy, viewMode, initActionId, editActionId);
                
                if (result.containsKey("contentIds"))
                {
                    @SuppressWarnings("unchecked")
                    List<String> contentIds = (List<String>) result.getOrDefault("contentIds", new ArrayList<>());
                    for (String contentId : contentIds)
                    {
                        _initializeShareableFields(contentId, parentContentId, contentIds);
                    }
                }
            }
        }
        finally 
        {
            _solrIndexHelper.restartSolrCommitForEvents(handledEventIds);
        }
        
        return result;
    }
    
    /**
     * Initialize shareable fields for the copied content
     * @param copiedContentId the copied content id
     * @param parentContentId the parent content id
     * @param createdContentIds the list of created content ids by copy
     */
    protected void _initializeShareableFields(String copiedContentId, String parentContentId, List<String> createdContentIds)
    {
        Content mainContent = _resolver.resolveById(copiedContentId);
        if (mainContent instanceof Course)
        {
            // Get created parent course list of created course content
            Course course = (Course) mainContent;
            String courseListParentId = course.getParentCourseLists()
                .stream()
                .map(Content::getId)
                .filter(createdContentIds::contains) // filter on created course list by the copy
                .findFirst()
                .orElse(parentContentId);               
            
            UserIdentity user = _currentUserProvider.getUser();
            
            CourseList courseListParent = StringUtils.isNotBlank(courseListParentId) ? _resolver.resolveById(courseListParentId) : null;
            if (_shareableCourseHelper.initializeShareableFields(course, courseListParent, user, false))
            {
                course.saveChanges();
                
                // Create a new version
                course.checkpoint();
            }
        }
    }
    
    /**
     * Check that duplication can be performed without blocking errors
     * @param contentId The content id to copy
     * @param parentContentId The parent content id
     * @param duplicationMode The duplication mode
     * @param results the results map
     * @return true if the duplication can be performed
     */
    protected boolean checkBeforeDuplication(String contentId, String parentContentId, DuplicationMode duplicationMode, Map<String, Object> results)
    {
        boolean allRight = true;
        if (StringUtils.isNotBlank(parentContentId))
        {
            // Check if the parent is locked
            Content parentContent = _resolver.resolveById(parentContentId);
            if (_isLocked(parentContent))
            {
                @SuppressWarnings("unchecked")
                List<Map<String, Object>> lockedContents = (List<Map<String, Object>>) results.get("locked-contents");
                Map<String, Object> contentParams = getContentDefaultParameters (parentContent);
                contentParams.put("description", _getLockedDescription(parentContent));
                lockedContents.add(contentParams);
            }
        }
            
        Content content = _resolver.resolveById(contentId);
        if (content instanceof ProgramItem programItem)
        {
            if (duplicationMode == DuplicationMode.SINGLE)
            {
                // Check if the child are locked
                for (ProgramItem programItemChild : _odfHelper.getChildProgramItems(programItem))
                {
                    if (_isLocked((Content) programItemChild))
                    {
                        @SuppressWarnings("unchecked")
                        List<Map<String, Object>> lockedContents = (List<Map<String, Object>>) results.get("locked-contents");
                        Map<String, Object> contentParams = getContentDefaultParameters ((Content) programItemChild);
                        contentParams.put("description", _getLockedDescription((Content) programItemChild));
                        lockedContents.add(contentParams);
                        
                        allRight = false;
                    }
                }
            }
            else if (duplicationMode == DuplicationMode.STRUCTURE_ONLY)
            {
                // Check if course child are locked
                for (Course course : _getCourse(programItem))
                {
                    if (_isLocked(course))
                    {
                        @SuppressWarnings("unchecked")
                        List<Map<String, Object>> lockedContents = (List<Map<String, Object>>) results.get("locked-contents");
                        Map<String, Object> contentParams = getContentDefaultParameters (course);
                        contentParams.put("description", _getLockedDescription(course));
                        lockedContents.add(contentParams);
                        
                        allRight = false;
                    }
                }
            }
        }
        
        return allRight;
    }
    
    /**
     * Get all first courses in sub item of the program item
     * @param programItem the program item
     * @return a set of courses
     */
    protected Set<Course> _getCourse(ProgramItem programItem)
    {
        Set<Course> courses = new HashSet<>();
        for (ProgramItem programItemChild : _odfHelper.getChildProgramItems(programItem))
        {
            if (programItemChild instanceof Course)
            {
                courses.add((Course) programItemChild);
            }
            else
            {
                courses.addAll(_getCourse(programItemChild));
            }
        }
        
        return courses;
    }

    /**
     * Enumeration for the mode of duplication
     */
    public enum DuplicationMode
    {
        /** Duplicate the content only */
        SINGLE,
        /** Duplicate the content and its structure */
        STRUCTURE_ONLY,
        /** Duplicate the content and its structure and its courses */
        FULL
    }
}
